"""AGM API module."""

import json
from datetime import datetime
from urllib import request

from .cache import AGM_CACHE
from .results import AGMResults


AGM_URL_ENDPOINTS = {
    'JUICE_API': 'https://juicept.esac.esa.int/agm',
}


def agm_api(metakernel, ptr, url, cache=True):
    """Call AGM through a REST API.

    Parameters
    ----------
    metakernel: str
        Metakernel id.
    ptr: str
        Pointing Timeline Request.
    url: str, optional
        Explicit AGM URL endpoint. You can also use the mission
        name listed in ``AGM_URL_ENDPOINTS`` like ``'JUICE_API'``.
    cache: bool, optional
        Use cache the response if present locally.

    Raises
    ------
    ValueError
        If the URL endpoint is unknown.

    """
    if url in AGM_URL_ENDPOINTS:
        endpoint = AGM_URL_ENDPOINTS[url]
    elif '://' in url:
        endpoint = url
    else:
        raise ValueError(f'Unknown URL: {url}')

    fname = AGM_CACHE(metakernel, ptr, endpoint)

    if cache and fname.exists():
        return AGMResults(fname)

    payload = json.dumps({
        'metakernel': metakernel,
        'ptr_content': str(ptr),
    }).encode('utf-8')

    req = request.Request(endpoint, data=payload)

    with request.urlopen(req) as resp:
        results = {
            'endpoint': endpoint,
            'metakernel': metakernel,
            'ptr': str(ptr).strip(),
            'results': json.load(resp),
            'created': datetime.now().isoformat(timespec='seconds'),
            'cache': {
                'location': str(AGM_CACHE),
                'md5_hash': fname.stem,
            } if cache else False,
        }

    if cache:
        with fname.open('w', encoding='utf-8') as f:
            json.dump(results, f)

    return AGMResults(results)



"""PTR Elements module."""

from functools import wraps
from pathlib import Path
from xml.dom import getDOMImplementation

from .datetime.parser import dt, td


XML = getDOMImplementation().createDocument(None, None, None)


class FrozenElementError(Exception):
    """Frozen element error."""


def frozen_lock(func):
    """Freezing decorator."""
    @wraps(func)
    def wrap(self, *args, **kwargs):
        """Wrapper function."""
        if self.is_frozen():
            raise FrozenElementError(f'<{self.tag}> cannot be changed.')
        return func(self, *args, **kwargs)
    return wrap


class Element:
    """Abstract PTR element object.

    Parameters
    ----------
    tag: str
        Element tag name.
    *elements: any
        Text/numerical values or list of children elements.
    description: str or list, optional
        Element description, put as a xml-comment on top of the element.
    **attrs: str
        Element attributes.

    """
    INDENT = '  '
    LIST_SEP = '  '

    def __init__(self, tag, *elements, description=None, **attrs):
        self.tag = tag
        self.description = description
        self.attrs = attrs

        for element in elements:
            self.append(element)

    def __str__(self):
        s = ''
        for desc in self.xml_desc:
            s += desc.toprettyxml(indent=self.INDENT)

        s += self.xml.toprettyxml(indent=self.INDENT)
        return s.strip()

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return str(self) == str(other)

    def __add__(self, other):
        return self.append(other)

    def __len__(self):
        return len(self._els)

    def __iter__(self):
        return iter(self._els)

    def __getitem__(self, key):
        if isinstance(key, int):
            return self._els[key]

        elements = []
        for el in self:
            if isinstance(el, Element):
                if el.tag == key:
                    elements.append(el)
                elif key in el:
                    children = el[key]
                    if isinstance(children, list):
                        elements.extend(children)
                    else:
                        elements.append(children)

        if not elements:
            raise KeyError(key)

        return elements[0] if len(elements) == 1 else elements

    def __setitem__(self, key, value):
        self[key].set_value(value)

    def __contains__(self, key):
        for el in self:
            if isinstance(el, Element) and (el.tag == key or key in el):
                return True
        return False

    def __delitem__(self, key):
        self.pop(key)

    @property
    def tag(self):
        """Element tag name."""
        return self._tag

    @tag.setter
    def tag(self, tag):
        """Element tag name setter."""
        self._tag = tag
        self._els = []
        self._attrs = {}
        self._desc = []

    @property
    def value(self):
        """Element parsed value."""
        if len(self) != 1 or isinstance(self[0], Element):
            raise ValueError('Only single value can be retrieved with `.value`')

        return self._value_parser(self[0])

    def _value_parser(self, value):
        """Value parser."""
        if self.LIST_SEP in str(value):
            return [
                self._value_parser(v.strip())
                for v in str(value).split(self.LIST_SEP)
                if v.strip()
            ]

        if str(value).isdecimal():
            return int(value)

        for func in [float, dt, td]:
            try:
                return func(value)
            except (ValueError, TypeError):
                pass

        return value

    def set_value(self, value):
        """Set element value."""
        if len(self) != 1 or isinstance(self[0], Element):
            raise ValueError('Only single value can be edited.')

        self._els[0] = value

        return self

    def freeze(self, status=True):
        """Freeze the element."""
        setattr(self, 'frozen', status)

    def is_frozen(self):
        """Check if the element is frozen."""
        return getattr(self, 'frozen', False)

    @frozen_lock
    def append(self, element):
        """Append a new element or text/numeric value."""
        if len(self) == 1 and not isinstance(self[0], Element):
            raise ValueError(f'<{self.tag}> already content the value '
                             f'`{self[0]}` impossible to add `{element}`.')

        if isinstance(element, dict):
            for tag, value in element.items():
                self.append(Element(tag, value))
            return self

        if self and not isinstance(element, Element):
            raise ValueError(
                f'<{self.tag}> already content {len(self)} `ptr.Element`, '
                f'impossible to add `{element}`.')

        if element is not None:
            self._els.append(element)

        return self

    @property
    def description(self):
        """Element description."""
        return self._desc

    @description.setter
    def description(self, desc):
        """Element description setter."""
        if not desc:
            self._desc = []
        elif isinstance(desc, (list, tuple)):
            self._desc.extend(d for d in desc)
        else:
            self._desc.append(desc)

    @property
    def xml_desc(self):
        """Element XML formatted description."""
        return (
            XML.createComment(f' {desc.replace("-", "–").strip()} ')
            for desc in self.description
        )

    @property
    def xml(self):
        """Element XML object."""
        xml = XML.createElement(self.tag)

        # Add element attributes
        for key, value in self.attrs.items():
            if value is not None:
                xml.setAttribute(key, str(value))

        for element in self:
            if isinstance(element, Element):
                # Add child description if present
                for desc in element.xml_desc:
                    xml.appendChild(desc)

                child = element.xml
            else:
                child = XML.createTextNode(f' {self._fmt_el(element)} ')

            xml.appendChild(child)

        return xml

    def _fmt_el(self, element) -> str:
        """Format element as XML text string."""
        if not isinstance(element, str) and hasattr(element, '__iter__'):
            return self.LIST_SEP.join([
                self._fmt_el(el)
                for el in element
            ])

        return str(element).strip()

    def index(self, tag) -> int:
        """Locate tag element index in the elements list.

        Parameters
        ----------
        tag: str
            Tag element name to search in the Element.

        Returns
        -------
        int
            Index of the tag element requested.

        Raises
        ------
        ValueError
            If the tag element was not found in the Element.

        Note
        ----
        If multiple tag elements are present only the index
        of the first tag element found is returned.
        Only the top level children are explored.

        """
        for i, el in enumerate(self):
            if el.tag == tag:
                return i

        raise ValueError(f'`{tag}` is not present in {self}')

    @frozen_lock
    def pop(self, key=-1):
        """Pop child element(s) by index or key.

        Parameters
        ----------
        key: int or str
            Element index or Element tag name to remove.
            By default the last element is removed.

        Returns
        -------
        Element or list(Element, ...)
            Popped element(s).

        Raises
        ------
        IndexError
            If the provided index is out of range.
        KeyError
            If not element match the provided key.
        TypeError
            If the key provided is invalid.

        Note
        ----
        When removed by tag name, if multiple child elements match
        the provided key, they will all be removed and returned.

        """
        if isinstance(key, int):
            return self._els.pop(key)

        if isinstance(key, str):
            if key not in self:
                raise KeyError(f'`{key}` is not in `{self.__class__.__name__}`.')

            els, ipop = [], []
            for i, el in enumerate(self._els):
                if isinstance(el, Element):
                    if el.tag == key:
                        els.append(el)  # Store the element to remove
                        ipop.append(i)    # and its index

                    elif key in el:
                        p = el.pop(key)
                        if isinstance(p, list):
                            els.extend(p)
                        else:
                            els.append(p)

            for i in reversed(ipop):
                self._els.pop(i)  # Remove the els elements here

            return els[0] if len(els) == 1 else els

        raise TypeError(
            f'Only `int` and `str` are accepted (`{type(key).__name__}` provided).')

    @frozen_lock
    def insert(self, index, element):
        """Insert element before a given index."""
        self.append(element)         # Check if the element can be append
        el = self.pop()              # Remove the last added element
        self._els.insert(index, el)  # Insert the element at the right position

    def save(self, filename, overwrite=False):
        """Save the element into a file.

        Parameters
        ----------
        filename: str or pathlib.Path
            Output filename.
        overwrite: bool, optional
            Overwrite the file if already exists (default: False).

        Returns
        -------
        pathlib.Path
            Output file location.

        """
        fname = Path(filename)

        if fname.exists() and not overwrite:
            raise FileExistsError(filename)

        fname.write_text(str(self) + '\n', encoding='utf-8')

        return fname



"""ESA Planning Timeline Request package."""

from importlib.metadata import version

from .agm import agm_simulation
from .block import ObsBlock
from .datetime import Time
from .element import Element
from .prm import PointingRequestMessage
from .ptx import read_ptr, read_ptx


__all__ = [
    'Time',
    'Element',
    'ObsBlock',
    'PointingRequestMessage',
    'read_ptx',
    'read_ptr',
    'agm_simulation',
]

__version__ = version('esa-ptr')



"""HTML helpers."""

import uuid


COLLAPSIBLE_STYLE = '''
ul.collapsible {
  list-style: none;
  margin: 0;
  padding-left: 0 !important;
}
ul.collapsible li input + label {
  display: inline-block;
  width: 140px;
  color: #777;
  font-weight: 500;
  padding: 4px 0 2px 0;
}
ul.collapsible li input:enabled + label {
  cursor: pointer;
}
ul.collapsible li input[type="checkbox"] {
  display: none;
}
ul.collapsible li input + label:before {
  content: '► ';
  font-size: 11px;
  color: #777;
  width: 20px;
}
ul.collapsible li input:checked + label:before {
  content: '▼ ';
}
ul.collapsible li input ~ div {
  display: none;
}
ul.collapsible li input:checked ~ div {
  display: block;
}
ul.collapsible ul.collapsible {
  list-style: none;
  margin-left: 1em;
}
'''


class HTMLCollapsible:
    """HTML collapsible element.

    Parameters
    ----------
    *elements: tuple
        Elements to put in the collapsible list.
        They must be formatted as:

        ``('header-1', data_1), ('header-2', data_2, 'expand'), …``

        The keyword `'expand'`` is optional, if provided
        the list will be expanded.

    Note
    ----
    If the provided data element has a `_repr_html_()` representation,
    if will be forwarded in the list representation.

    """

    def __init__(self, *elements):
        self.html = f'<style>{COLLAPSIBLE_STYLE}</style>\n'

        self.html += '<ul class="collapsible">\n'
        for header, data, *expand in elements:
            _id = uuid.uuid4()

            if expand and 'expand' in str(expand).lower():
                checked = 'checked'
            else:
                checked = 'unchecked'

            # Get HTML representation
            if hasattr(data, '_repr_html_'):
                content = html(data)

            elif isinstance(data, list):
                content = '<ul>\n<li>'
                content += '</li>\n<li>'.join([f'{d}' for d in data])
                content += '</li>\n</ul>'

            elif isinstance(data, dict):
                content = '<ul>\n<li>'
                content += '</li>\n<li>'.join([
                    f'<b>{key}</b> {value}' for key, value in data.items()
                ])
                content += '</li>\n</ul>'

            else:
                content = f'{data}'

            self.html += (
                '<li>\n'
                f'    <input id="{_id}" type="checkbox" {checked}>\n'
                f'    <label for="{_id}">{header}</label>\n'
                f'    <div>{content}</div>\n'
                '</li>\n'
            )

        self.html += '</ul>'

    def __str__(self):
        return self.html

    def __repr__(self):
        return str(self)

    def _repr_html_(self):
        return str(self)


def html(obj):
    """Return object _repr_html_() representation."""
    if repr_html := getattr(obj, '_repr_html_', False):
        return repr_html()

    raise NotImplementedError(f'`{obj.__class__.__name__}` has no HTML representation.')



"""Generic PTR snippets."""

from .element import Element


class Lon(Element):
    """PTR east longitude element.

    Parameters
    ----------
    lon: float
        East longitude value.
    units: str, optional
        Value units (default: 'deg').
    description: str or list, optional
        Longitude description, put as a xml-comment on top of the element.

    """
    def __init__(self, lon_e, units='deg', description=None):
        super().__init__('lon', float(lon_e), units=units, description=description)


class Lat(Element):
    """PTR latitude element.

    Parameters
    ----------
    lon: float
        Latitude value.
    units: str, optional
        Value units (default: 'deg').
    description: str or list, optional
        Latitude description, put as a xml-comment on top of the element.

    """
    def __init__(self, lat, units='deg', description=None):
        super().__init__('lat', float(lat), units=units, description=description)


class Target(Element):
    """PTR target element.

    Parameters
    ----------
    lon_e: float
        Target east longitude coordinate.
    lat: float
        Target latitude coordinate.
    frame: str, optional
        Target reference frame (default 'EME2000').
    units: str, optional
        Value coordinates units (default: 'deg').
    description: str or list, optional
        Target description, put as a xml-comment on top of the element.
    **attrs: optional
        Targets attributes.

    """
    def __init__(self, lon_e, lat, frame='EME2000', units='deg', **attrs):
        super().__init__(
            'target',
            Lon(lon_e, units=units),
            Lat(lat, units=units),
            frame=frame,
            **attrs
        )


class SolarArrays(Element):
    """PTR metadata solar array element.

    Parameters
    ----------
    fixed_rotation_angle: float, optional
        Solar array fixed rotation angle (in degrees).
    description: str or list, optional
        Solar array description, put as a xml-comment on top of the element.

    """

    def __init__(self, fixed_rotation_angle=None, description=None, **attrs):
        super().__init__('solarArrays', description=description, **attrs)

        if fixed_rotation_angle is not None:
            self.append(
                Element('fixedRotationAngle', fixed_rotation_angle, units='deg')
            )



"""PTR datetime window module."""

from functools import wraps

from .parser import dt
from .time import EndTime, StartTime
from ..element import Element


def is_window(func):
    """Check if the other is derived from a datetime window."""
    @wraps(func)
    def wrap(self, other):
        """Wrapped function."""
        if not isinstance(other, WithDatetimeWindow):
            raise TypeError(f'{self.__class__.__name__} can not be compared to '
                            f'{other.__class__.__name__}.')
        return func(self, other)
    return wrap


class WithDatetimeWindow:
    """Add Time window properties to Element objects."""
    def __init__(self):
        self._start, self._end = None, None  # pragma: no cover

    @is_window
    def __and__(self, other) -> bool:
        """Check intersection with an other datetime window.

        ----------|◁◁◁ self  ▷▷▷|----------
        ----░░░---|--- other ---|---░░░----
        -------░░░|-------------|░░░-------
        -------███|███-------███|███-------
        ----------|███-------███|----------
        ----------|█████████████|███-------
        -------███|█████████████|----------
        ----------|---███████---|----------
        ----------|█████████████|----------
        -------███|█████████████|███-------

        self & other == True: █ / False: ░

        """
        return self.start < other.end and self.end > other.start

    @is_window
    def __lt__(self, other) -> bool:
        """Exclusive lower datetime window comparison.

        Self starts before and is shorter than other:

        ----------|◁◁◁ self  ▷▷▷|----------
        ----░░░---|--- other ---|---███----
        -------░░░|-------------|███-------
        -------░░░|░░░-------███|███-------
        ----------|░░░-------███|----------
        ----------|█████████████|███-------
        -------░░░|░░░░░░░░░░░░░|----------
        ----------|---███████---|----------
        ----------|░░░░░░░░░░░░░|----------
        -------░░░|░░░░░░░░░░░░░|░░░-------

        self < other == True: █ / False: ░

        """
        return self.start < other.start or \
            (self.start == other.start and self.end < other.end)

    @property
    def start(self) -> StartTime:
        """Time window start time value."""
        return self._start

    @property
    def end(self) -> EndTime:
        """Time window end time value."""
        return self._end

    @property
    def window(self):
        """Time window start and end time values."""
        return self.start, self.end

    def set_window(self, start, end):
        """Set time window boundaries."""
        if hasattr(self, '_start') or hasattr(self, '_end'):
            raise AttributeError('Use `.edit()` method instead.')

        start_time = StartTime(start)
        end_time = EndTime(end)

        if start_time >= end_time:
            raise ValueError('Start time shall be before end time.')

        self._start, self._end = start_time, end_time

        self.append(self._start)
        self.append(self._end)

    @property
    def duration(self):
        """Block duration."""
        return self.end - self.start

    def append(self, element):
        """Append a new element or text/numeric value."""
        raise NotImplementedError

    def edit(self, *, start=None, end=None):
        """Edit temporal window boundaries.

        Parameters
        ----------
        start: str, datetime.datetime or datetime.timedelta
            Start time absolute or relative offset.

        end: str, datetime.datetime or datetime.timedelta
            End time absolute or relative offset.

        Raises
        ------
        ValueError
            If the new start time is after the end time.
        ValueError
            If the new end time is before the start time.

        Warning
        -------
        This operation change the duration of the temporal window.

        """
        if start is None:
            start_dt = self.start.datetime
        else:
            try:
                start_dt = self.start + start  # Relative offset (timedelta)
            except ValueError:
                start_dt = dt(start)           # Absolute offset (datetime)

        if end is None:
            end_dt = self.end.datetime
        else:
            try:
                end_dt = self.end + end  # Relative offset (timedelta)
            except ValueError:
                end_dt = dt(end)         # Absolute offset (datetime)

        # Check that the new window boundaries are still valid
        if end_dt <= start_dt:
            raise ValueError(
                f'Start time `{start_dt.isoformat()}` must be before '
                f'end time: `{end_dt.isoformat()}`.')

        # Edit window only if required
        if start:
            self.start.edit(start_dt)

        if end:
            self.end.edit(end_dt)

        return self

    def offset(self, offset, *, ref='start'):
        """Offset the temporal window globally.

        Parameters
        ----------
        offset: str datetime.timedelta or datetime.datetime
            Global or relative offset.

        ref: str, optional
            Boundary reference for relative offset.
            Only ``start|end|center`` are accepted

        Raise
        -----
        KeyError
            If the reference keyword is invalid.

        Note
        ----
        This operation does not change the duration of the window.

        """
        try:
            # Global offset (with timedelta)
            self.start.offset(offset)
            self.end.offset(offset)

        except ValueError:
            # Relative offset (with datetime)
            t, d = dt(offset), self.duration

            if ref == 'start':
                self.start.edit(t)
                self.end.edit(t + d)

            elif ref == 'end':
                self.start.edit(t - d)
                self.end.edit(t)

            elif ref == 'center':
                self.start.edit(t - d / 2)
                self.end.edit(t + d / 2)

            else:
                raise KeyError('For relative offset, the ref keyword must be '
                               'in `start|end|center`.') from None
        return self


class ElementWindow(Element, WithDatetimeWindow):
    """Element with datetime window properties."""

    def __init__(self, tag, start, end, *elements, **attrs):
        super().__init__(tag, **attrs)
        self.set_window(start, end)

        for element in elements:
            self.append(element)



"""AGM cache module."""

import re
from pathlib import Path
from tempfile import gettempdir

from .hash import md5_hash
from .results import AGMResults


class Cache:
    """AGM cache folder.

    Parameters
    ----------
    cache: str or pathlib.Path
        Cache folder name.

    """

    def __init__(self, cache):
        self.cache = Path(cache)
        self.cache.mkdir(parents=True, exist_ok=True)

    def __str__(self):
        return str(self.cache)

    def __repr__(self):
        n = len(self)

        if n == 0:
            n_files = 'empty'
        else:
            n_files = f'{n} entr' + ('ies' if n > 1 else 'y') + ':\n - '
            n_files += '\n - '.join([
                f.stem for f in self.content
            ])

        return f'<{self.__class__.__name__}> {self} | {n_files}'

    def __call__(self, metakernel, ptr, endpoint, ext='json'):
        return self.cache / f'{md5_hash(metakernel, ptr, endpoint)}.{ext}'

    def __len__(self):
        return len(self.content)

    def __iter__(self):
        return iter(self.content)

    def __contains__(self, key):
        return (self.cache / f'{key}.json').exists()

    def __getitem__(self, key):
        if key not in self:
            raise KeyError(key)
        return self.inspect(key)

    def __delitem__(self, key):
        if key not in self:
            raise KeyError(key)
        self.remove(key)

    @property
    def content(self):
        """List CK and NPY files in AGM cache"""
        return sorted(
            f for f in self.cache.glob('*.json')
            if re.match(r'^[a-f\d]{32}$', f.stem)
        )

    def inspect(self, md5):
        """Inspect cached file."""
        return AGMResults(self.cache / f'{md5}.json')

    def remove(self, md5):
        """Remove the file from the cache."""
        (self.cache / f'{md5}.json').unlink(missing_ok=True)
        (self.cache / f'{md5}.ck').unlink(missing_ok=True)

    def purge(self):
        """Purge all AGM cache"""
        for fname in self:
            self.remove(fname.stem)


AGM_CACHE = Cache(Path(gettempdir()) / 'esa-ptr')



"""PTR time module."""

from datetime import datetime

from .parser import dt, td
from ..element import XML, Element


class WithTime:
    """Add Time properties to Element objects."""

    def __add__(self, other):
        return self.datetime + td(other)  # -> datetime

    def __sub__(self, other):
        try:
            return self.datetime - dt(other)  # -> timedelta
        except ValueError:
            return self.datetime - td(other)  # -> datetime

    def __gt__(self, other) -> bool:
        return (self - dt(other)).total_seconds() > 0

    def __lt__(self, other) -> bool:
        return (self - dt(other)).total_seconds() < 0

    def __ge__(self, other) -> bool:
        return (self - dt(other)).total_seconds() >= 0

    def __le__(self, other) -> bool:
        return (self - dt(other)).total_seconds() <= 0

    @property
    def datetime(self) -> datetime:
        """Datetime value."""
        raise NotImplementedError

    def edit(self, value):
        """Edit time value."""
        raise NotImplementedError

    @property
    def date(self):
        """Date value."""
        return self.datetime.date()

    @property
    def time(self):
        """Time value."""
        return self.datetime.time()

    @property
    def iso(self):
        """ISO formatted datetime."""
        return self.datetime.isoformat()

    def item(self) -> datetime:
        """Datetime value (numpy-like)."""
        return self.datetime

    def offset(self, offset):
        """Shift time value."""
        self.edit(self.datetime + td(offset))
        return self


class Time(WithTime):
    """Generic Time object."""
    def __init__(self, time):
        self._dt = dt(time)

    def __str__(self):
        return self.iso

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.iso}')"

    def __eq__(self, other) -> bool:
        return self.datetime == dt(other)

    @property
    def datetime(self) -> datetime:
        """Datetime value."""
        return self._dt

    def edit(self, value):
        """Edit time value."""
        self._dt = dt(value)
        return self


class ElementTime(WithTime, Element):
    """Time element.

    Parameters
    ----------
    tag: str
        Element tag name.
    time: str, datetime.datetime or numpy.datetime64
        Element time.
    description: str or list, optional
        Element description, put as a xml-comment on top of the element.

    Note
    ----
    The element is parsed as a native datetime.datetime object,
    and display in XML in ISO format: ``YYYY-MM-DDThh:mm:ss``.

    """
    def __init__(self, tag, time, description=None):
        if not description and isinstance(time, ElementTime):
            description = time.description

        super().__init__(tag, dt(time), description=description)

    @property
    def datetime(self) -> datetime:
        """Datetime value."""
        return self[0]

    def edit(self, value):
        """Edit time value."""
        self._els[0] = dt(value)
        return self

    @property
    def xml(self):
        """XML PTR time object."""
        xml = XML.createElement(self.tag)
        xml.appendChild(XML.createTextNode(f' {self.iso} '))
        return xml


class StartTime(ElementTime):
    """PTR start ElementTime element.

    Parameters
    ----------
    time: str, datetime.datetime or numpy.datetime64
        Element start time.
    description: str or list, optional
        Element description, put as a xml-comment on top of the element.

    """
    def __init__(self, time, description=None):
        super().__init__('startTime', time, description=description)


class EndTime(ElementTime):
    """PTR end time element.

    Parameters
    ----------
    time: str, datetime.datetime or numpy.datetime64
        Element end time.
    description: str or list, optional
        Element description, put as a xml-comment on top of the element.

    """
    def __init__(self, time, description=None):
        super().__init__('endTime', time, description=description)



"""PTR axis module."""

from .element import Element


AXIS = {
    'X': [1, 0, 0],
    'Y': [0, 1, 0],
    'Z': [0, 0, 1],
}


def invert(vec):
    """Invert vector values."""
    return [-1 * v for v in vec]


def parse_vector(*axis):
    """Parse axis vector.

    Parameters
    ----------
    *axis: tuple or str
        Axis vector tuple or name (eg. ``(1, 0, 0)`` or '-Y').

    Returns
    -------
    list
        X, Y, Z vector components.

    """
    if len(axis) == 1 and isinstance(axis[0], str):
        name = axis[0]
        vec = name.upper()[-1]

        if vec not in ['X', 'Y', 'Z']:
            raise ValueError('Only (+/-) X, Y, Z are considered as valid.')

        vec = AXIS[vec]

        if name.startswith('-'):
            vec = invert(vec)

    elif len(axis) == 3:
        vec = list(axis)

    else:
        raise ValueError('This axis vector must have 3 components (x, y, z).')

    return vec


class ElementAxis(Element):
    """PTR axis element.

    Parameters
    ----------
    frame: str
        Element reference frame.
    *axis: tuple or str
        Axis vector tuple or axis name (eg. '-Y').
    description: str or list, optional
        Axis description, put as a xml-comment on top of the element.

    """

    def __init__(self, frame, *axis, description=None):
        self.vec = parse_vector(*axis)

        super().__init__(
            f'{frame}Axis',
            {'x': self.x, 'y': self.y, 'z': self.z},
            frame=frame,
            description=description,
        )

    @property
    def frame(self):
        """Axis frame attribute."""
        return self.attrs['frame']

    def __neg__(self):
        return ElementAxis(self.frame, *invert(self.vec))

    def __pos__(self):
        return ElementAxis(self.frame, *self.vec)

    @property
    def x(self):
        """X-axis value."""
        return self.vec[0]

    @property
    def y(self):
        """Y-axis value."""
        return self.vec[1]

    @property
    def z(self):
        """Z-axis value."""
        return self.vec[2]


class ScAxis(ElementAxis):
    """PTR spacecraft axis element.

    Parameters
    ----------
    *xyz: tuple or str
        Spacecraft vector tuple or axis name (eg. ``(1, 0, 0)`` or '-Y').
    description: str or list, optional
        Axis description, put as a xml-comment on top of the element.

    """

    def __init__(self, *axis, description=None):
        super().__init__('SC', *axis, description=description)


# Spacecraft axis shortcuts
xScAxis = ScAxis('X')
yScAxis = ScAxis('Y')
zScAxis = ScAxis('Z')



"""PTR time module."""

import re
from datetime import date, datetime, timedelta


def dt(time) -> datetime:
    """Datetime parser."""
    if isinstance(time, datetime):
        return time

    if isinstance(time, date):
        return datetime(time.year, time.month, time.day)

    if hasattr(time, 'item'):
        return time.item()  # numpy.datetime64

    t = str(time)

    if t.endswith('.'):
        t = t[:-1]  # Remove the dot if present at the end

    return datetime.fromisoformat(t)


STD_DURATION = re.compile(
    r'^'
    r'(?:(?P<day>-?\d+) (days?,? )?)?'
    r'((?:(?P<hr>-?\d+):)(?=\d+:\d+))?'
    r'(?:(?P<min>-?\d+):)?'
    r'(?P<sec>-?\d+(?:\.\d+)?)'
    r'$'
)
UNITS_DURATION = re.compile(
    r'^'
    r'(?:(?P<day>[+-]?\d+(?:\.\d+)?)\s*(?:d|D|days?))?\s*'
    r'(?:(?P<hr>[+-]?\d+(?:\.\d+)?)\s*(?:h|hrs?|hours?))?\s*'
    r'(?:(?P<min>[+-]?\d+(?:\.\d+)?)\s*(?:m|mins?|minutes?))?\s*'
    r'(?:(?P<sec>[+-]?\d+(?:\.\d+)?)\s*(?:s|secs?|seconds?))?\s*'
    r'(?:(?P<ms>[+-]?\d+(?:\.\d+)?)\s*(?:ms|msecs?|milliseconds?))?'
    r'$'
)
MICRO_SEC = {
    'ms': 1e-3, 'sec': 1, 'min': 60, 'hr': 3_600, 'day': 86_400,
}


def td(delta) -> timedelta:
    """Time delta parser.

    Example: ``7 day, 12:34:56.789`` for
    7 days, 12 hours, 34 minutes, 56 secondes and 789 milliseconds.

    We recommend to use this formatting but you can provide
    explicit units following the rules below:

    Valid units:

    - ``ms``, ``msec``, ``millisecond``
    - ``s``, ``sec``, ``second``
    - ``m``, ``min``, ``minute``
    - ``h``, ``hr``, ``hour``
    - ``d``, ``D``, ``day``

    Plural units are also valid.

    Space between the value and the units is accepted.

    Month(s) and year(s) are not accepted.

    If a ``int`` or a ``float`` is provide, it is assume
    that the value is express in secondes.

    Raises
    ------
    ValueError
        If the provided value format is invalid.

    """
    if isinstance(delta, timedelta):
        return delta

    s = str(delta)

    match = STD_DURATION.match(s) or UNITS_DURATION.match(s)

    if not match:
        raise ValueError(f'Invalid format: `{delta}`')

    us = 1e6 * sum(
        float(value) * MICRO_SEC[unit]
        for unit, value in match.groupdict(default=0).items()
    )

    return timedelta(microseconds=us)



"""PTR blocks module."""

from ..datetime.window import WithDatetimeWindow
from ..element import Element
from ..metadata import WithMetadata


class ElementBlock(Element, WithDatetimeWindow, WithMetadata):
    """PTR block element.

    Parameters
    ----------
    ref: str
        Block reference attribute.
    start: string, datetime.datetime or numpy.datetime64
        Block start time.
    end: string, datetime.datetime or numpy.datetime64
        Block end time.
    *elements:
        Other block elements.

    metadata: str, int, float, ptr.Element, tuple or list, optional
        Block metadata element(s).
    description: str or list, optional
        Block description, put as a xml-comment on top of the element.
    **attrs:
        Block keywords attributes.

    """

    def __init__(self, ref, start, end, *elements,
                 metadata=None, description=None, **attrs):
        super().__init__('block', ref=ref, description=description, **attrs)

        if metadata:
            self.add_metadata(metadata)

        self.set_window(start, end)

        for element in elements:
            self.append(element)

        # Freeze the top element list
        self.freeze()

        # Init parents timelines list
        self._timelines = []

    def link_timeline(self, timeline):
        """Link a parent timeline to a block."""
        self._timelines.append(timeline)

    def unlink_timeline(self, timeline):
        """Unlink a parent timeline to a block."""
        self._timelines.remove(timeline)

    def edit(self, *, start=None, end=None):
        """Edit block temporal window boundaries.

        Parameters
        ----------
        start: str, datetime.datetime or datetime.timedelta
            Start time absolute or relative offset.

        end: str, datetime.datetime or datetime.timedelta
            End time absolute or relative offset.

        Raises
        ------
        BlockOverlapError
            If the new block is not compatible with its
            parent timelines.

        Warning
        -------
        This operation change the duration of the block.

        """
        # Store original start and end times
        original_start, original_end = self.start.iso, self.end.iso

        # Edit times
        super().edit(start=start, end=end)

        # Tag the new block (to avoid block duplicates)
        self.attrs['block'] = 'new'

        # Check if the new block is compatible with their parent timeline
        for timeline in self._timelines:
            for block in timeline:
                if block != self and block & self:
                    # Untag the new block
                    del self.attrs['block']

                    # Set error message
                    err_msg = f'New block:\n{self}\n\nintersects:\n\n{block}'

                    # Fallback to the original datetime
                    super().edit(start=original_start, end=original_end)

                    # Raise overlap error
                    raise BlockOverlapError(err_msg)

            # Re-sort the blocks in chronological order
            timeline.sort()

        # Untag the new block
        del self.attrs['block']

        return self

    def offset(self, offset, *, ref='start'):
        """Offset block temporal window.

        Parameters
        ----------
        offset: str datetime.timedelta or datetime.datetime
            Global or relative offset.

        ref: str, optional
            Boundary reference for relative offset.
            Only ``start|end|center`` are accepted

        Raises
        ------
        BlockOverlapError
            If the new block is not compatible with its
            parent timelines.

        Note
        ----
        This operation does not change the duration of the block.

        """
        # Store original start and end times
        original_start, original_end = self.start.iso, self.end.iso

        # Offset times
        super().offset(offset, ref=ref)

        # Tag the new block
        self.attrs['block'] = 'new'

        # Check if the new block is compatible with their parent timeline
        for timeline in self._timelines:
            for block in timeline:
                if block != self and block & self:
                    # Untag the new block
                    del self.attrs['block']

                    # Set error message
                    err_msg = f'New block:\n{self}\n\nintersects:\n\n{block}'

                    # Fallback to the original datetime
                    super().edit(start=original_start, end=original_end)

                    # Raise overlap error
                    raise BlockOverlapError(err_msg)

            # Re-sort the blocks in chronological order
            timeline.sort()

        # Untag the new block
        del self.attrs['block']

        return self


class BlockOverlapError(Exception):
    """Block overlap error."""



"""ESA Attitude Generator Module (AGM)."""

from .cache import AGM_CACHE
from .results import AGMResults
from .simulation import agm_simulation


__all__ = [
    'agm_simulation',
    'AGMResults',
    'AGM_CACHE',
]



"""PTR SLEW block."""

from ..element import Element


# An attitude slew is implemented by inserting a slew block in the PTR.
# A slew block must be placed in between two observation blocks.
# The duration of slew blocks is defined implicitly by the end time
# of the previous observation block and the start time of the following
# observation block. A slew block has the syntax: `<block ref="SLEW" />`
SLEW_BLOCK = Element('block', ref='SLEW')
SLEW_BLOCK.freeze()



"""AGM hashing module."""

import json
from hashlib import md5


def md5_hash(metakernel: str, ptr: str, endpoint: str) -> str:
    """MD5 inputs hashing method."""
    return md5(json.dumps({
        'endpoint': endpoint,
        'metakernel': metakernel,
        'ptr': str(ptr).strip(),
    }).encode('utf-8')).hexdigest()



"""PTR Command Line Interface module."""

import sys
from argparse import ArgumentParser
from pathlib import Path

from .agm import agm_simulation


def cli_juice_agm(argv=None):
    """CLI JUICE PTR validation with AGM.

    By default, the metakernel used by AGM is ``juice_crema_5_0b23_1``.
    You can change this parameter with ``--mk`` flag.

    AGM produced CK file are cached but you can download
    it explicitly if you want with a ``--ck`` flag.

    """
    parser = ArgumentParser(description='Check ESA/JUICE PTR validity with AGM.')
    parser.add_argument('ptr', help='PTR/PTX input file.')
    parser.add_argument('--mk', metavar='METAKERNEL', default='juice_crema_5_0',
                        help='Metakernel to use with AGM (default: `juice_crema_5_0`).')
    parser.add_argument('--endpoint', metavar='URL', default='JUICE_API',
                        help='AGM API endpoint URL (default: `JUICE_API`).')
    parser.add_argument('--quaternions', action='store_true',
                        help='Display computed quaternions (optional).')
    parser.add_argument('--ck', metavar='FILENAME',
                        help='CK output filename (optional).')
    parser.add_argument('--log', action='store_true',
                        help='Display AGM log (optional).')
    parser.add_argument('--no-cache', action='store_false',
                        help='Disable cache (optional).')

    args, _ = parser.parse_known_args(argv)

    # check if the file exists
    if not Path(args.ptr).exists():
        sys.stderr.write(f'PTR not found: {args.ptr}\n')
        sys.exit(1)

    # Start AGM simulation
    res = agm_simulation(args.mk, args.ptr, args.endpoint, cache=args.no_cache)
    sys.stdout.write(f'AGM simulation: {res.status}\n')

    if res.success:
        if args.log:
            sys.stdout.write('Log:\n' + repr(res.log) + '\n')

        if args.quaternions:
            sys.stdout.write(f'Quaternions:\n{repr(res.quaternions)}\n')

        if args.ck:
            res.ck.save(args.ck)
            sys.stdout.write(f'AGM CK saved in `{args.ck}`.\n')
    else:
        sys.stderr.write('Log:\n' + repr(res.log) + '\n')
        sys.exit(1)



"""PTR blocks utils module."""

from functools import wraps

from .block import BlockOverlapError, ElementBlock


def is_block(func):
    """Check if an element is an element block."""
    @wraps(func)
    def wrap(self, element):
        """Wrapped function."""
        if not isinstance(element, ElementBlock):
            raise TypeError(f'`{element}` is not an ElementBlock.')
        return func(self, element)
    return wrap


def after(latest, new_block) -> bool:
    """Check if the new block is after the latest one."""
    if latest.end <= new_block.start:
        return True

    raise BlockOverlapError(f'New block:\n{new_block}\n\nshould start after:\n\n{latest}')


def before(first, new_block) -> bool:
    """Check if the new block is before the first one."""
    if new_block.end <= first.start:
        return True

    raise BlockOverlapError(f'New block:\n{new_block}\n\nshould end before:\n\n{first}')


def between(block_1, block_2, new_block) -> bool:
    """Check if the new block is between 2 blocks."""
    return after(block_1, new_block) and before(block_2, new_block)


def gap(block_1, block_2) -> bool:
    """Check if 2 blocks have a gap between them."""
    return block_1.end < block_2.start or block_1.start > block_2.end


def insort(blocks: list, new_block):
    """Insert a new block in chronological order without overlap."""
    if not blocks:
        blocks.append(new_block)

    elif blocks[-1].start <= new_block.start and after(blocks[-1], new_block):
        blocks.append(new_block)

    else:
        for i in range(len(blocks) - 2, -1, -1):
            if blocks[i].start <= new_block.start:
                if between(blocks[i], blocks[i + 1], new_block):
                    blocks.insert(i + 1, new_block)
                    break
        else:
            if before(blocks[0], new_block):
                blocks.insert(0, new_block)



"""PTR observation blocks module."""

from .block import ElementBlock


class ObsBlock(ElementBlock):
    """PTR observation block.

    Parameters
    ----------
    start: string, datetime.datetime or numpy.datetime64
        Block start time.
    end: string, datetime.datetime or numpy.datetime64
        Block end time.
    *elements:
        Block elements.

    metadata: str, int, float, ptr.Element, tuple or list, optional
        Block metadata comments.
    description: str or list, optional
        Block description, put as a xml-comment on top of the element.
    **attrs:
        Block keywords attributes.

    """

    def __init__(self, start, end, *elements, metadata=None, description=None, **attrs):
        super().__init__(
            'OBS', start, end, *elements,
            metadata=metadata, description=description, **attrs
        )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['ptr', 'ptr.agm', 'ptr.block', 'ptr.datetime']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['juice-agm = ptr.cli:cli_juice_agm']}

setup_kwargs = {
    'name': 'esa-ptr',
    'version': '1.1',
    'description': 'ESA Planning Timeline Request package',
    'long_description': 'ESA Planning Timeline Request (PTR) Python package\n==================================================\n\nSince the [Rosetta mission](https://www.esa.int/Science_Exploration/Space_Science/Rosetta),\nESA developed an XML-like syntax to create Planning Timeline Request (PTR) files.\nThese files allow the mission team member to design custom attitude spacecraft pointing.\nIt is readable by `AGM` and `MAPPS` softwares to detect spacecraft constrains violations,\npower conception and surface coverage. It can also be used to compute custom spacecraft\nattitude (quaterions and `ck`).\nThis format will be re-used for the [JUICE mission](https://sci.esa.int/web/juice),\nand can already be tested on the [JUICE pointing tool](https://juicept.esac.esa.int).\n\nThis python package implements an object oriented approach to help the creation and parsing\nof PTR files for the user, as well an interface to check JUICE PTR validity with AGM.\n\n> **Disclaimer:** This package is an early release and does not support all PTR implementations.\n> Please, open an issue to report any issue you may accounter.\n> Currently this tool in **beta stage, do not use it in critical environments**.\n\nA detailed documentation can be found [here](https://esa-ptr.readthedocs.io/).\n\nInstallation\n------------\n\nThis package is available on [PyPI](https://pypi.org/project/esa-ptr/) and could be installed like this:\n\n```bash\npython -m pip install esa-ptr\n```\n\nEven if this tool does not have any external dependencies, we recommend to use it in an isolated virtual environment (`venv` or `conda`).\n\n\nDevelopment and testing\n-----------------------\n\nIf you want contribute to the development and tests your changes before submitting a merge request,\nyou need to install [Poetry](https://python-poetry.org/docs/) and clone this repository:\n\n```bash\ngit clone https://juigitlab.esac.esa.int/python/ptr.git python-esa-ptr ;\ncd python-esa-ptr/\n```\n\nInstall the package and its dependencies:\n```\npoetry install\n```\n\nThen, after your edits, you need to check that both linters are happy:\n```bash\npoetry run flake8\npoetry run pylint ptr tests/\n```\n\nand all the tests passed:\n```bash\npoetry run pytest\n```\n\n\nDocumentation\n-------------\n* Rosetta Flight Dynamics: `RO-ESC-IF-5501_i3r4_RSGS_FD_ICD-2.pdf`\n',
    'author': 'Benoît Seignovert',
    'author_email': 'benoit.seignovert@univ-nantes.fr',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://juigitlab.esac.esa.int/python/ptr',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Planning timeline parser."""

from html.parser import HTMLParser
from pathlib import Path

from .block import ElementBlock, ObsBlock
from .datetime import EndTime, StartTime
from .element import Element
from .metadata import Comment, File, Metadata
from .prm import PointingRequestMessage
from .timeline import Timeline


ELEMENTS = {
    'startTime': StartTime,
    'endTime': EndTime,
    'comment': Comment,
    'metadata': Metadata,
    'timeline': Timeline,
}


def el_parser(element):
    """Read element and convert it if necessary.

    Warning
    -------
    Only the top level element is parsed, not its children.

    """
    if isinstance(element, Element):
        if element.tag in ELEMENTS:
            return ELEMENTS[element.tag](
                *element,
                description=element.description,
                **element.attrs,
            )

        if element.tag == 'include':
            return File(element.attrs['href'], description=element.description)

        if element.tag == 'block':
            return block_parser(element)

        if element.tag == 'prm':
            return prm_parser(element)

    return element


def _extract(element, key):
    """Extract a key from a element if when is not present."""
    try:
        return list(element.pop(key))
    except KeyError:
        return None


def block_parser(element):
    """Parse block element.

    Parameters
    ----------
    element: ptr.Element
        PTR block element.

    Returns
    -------
    ElementBlock, ObsBlock or None
        Parsed block element.

    Raises
    ------
    AttributeError
        If the block does not have a defined `href` attribute.

    Note
    ----
    SLEW blocks are discarded (appended automatically in Timeline).

    """
    ref = element.attrs.get('ref', '').upper()

    if not ref:
        raise AttributeError(f'Block without `ref` attribute:\n{element}')

    # Remove `ref` from element attributes
    del element.attrs['ref']

    # Skip slew block
    if ref == 'SLEW':
        return None

    # Check start time
    if 'startTime' not in element:
        raise KeyError(f'Block without `startTime` element:\n{element}')

    # Remove only the 1st direct child
    start = element.pop(element.index('startTime'))

    # Check end time
    if 'endTime' not in element:
        raise KeyError(f'Block without `endTime` element:\n{element}')

    # Remove only the 1st direct child
    end = element.pop(element.index('endTime'))

    # Extract if metadata if present
    metadata = _extract(element, 'metadata')

    if ref == 'OBS':
        return ObsBlock(
            start, end, *element,
            metadata=metadata,
            description=element.description,
            **element.attrs,
        )

    return ElementBlock(
        ref, start, end, *element,
        metadata=metadata,
        description=element.description,
        **element.attrs,
    )


def prm_parser(element):
    """Parse pointing request message element.

    Parameters
    ----------
    element: ptr.Element
        PTR pointing request message element.

    Returns
    -------
    PointingRequestMessage
        Parsed pointing request message element.

    """
    header = _extract(element, 'header')

    # Check is the segment is present
    if 'segment' not in element:
        raise KeyError(f'Pointing request message without `segment` element:\n{element}')

    segment = element.pop('segment')

    # Check is the timeline is present
    if 'timeline' not in segment:
        raise KeyError(f'Pointing request message without `timeline` segment:\n{segment}')

    timeline = segment.pop('timeline')

    seg_name = segment.attrs.get('name')
    seg_metadata = _extract(segment, 'metadata')

    return PointingRequestMessage(
        *timeline,
        header=header,
        description=element.description,
        seg_name=seg_name,
        seg_metadata=seg_metadata,
        **element.attrs,
    )


class PtxParser(HTMLParser):
    """PTX content parser."""

    def __init__(self):
        super().__init__()

        self.data = None
        self.elements = []
        self.desc = []

    @property
    def tag(self):
        """Get case-sensitive tag name."""
        tag = self.get_starttag_text()

        for c in '</>':
            tag = tag.replace(c, '')

        return tag.split()[0]

    def handle_starttag(self, _, attrs):
        """Handle tag opening."""
        el = Element(self.tag, **dict(attrs))

        if self.desc:
            el.description = self.desc
            self.desc = []

        self.elements.append(el)

    def handle_endtag(self, tag):
        """Handle tag closure."""
        last_tag = self.elements[-1].tag.lower()

        if tag != last_tag:
            raise KeyError(f'Closing </{tag}> tag before </{last_tag}>.')

        # Get the last element of the list and convert it if necessary
        el = el_parser(self.elements.pop())

        if el is not None:
            if self.elements:
                self.elements[-1].append(el)
            else:
                self.data = el

    def handle_startendtag(self, tag, attrs):
        """Handle empty tag."""
        self.handle_starttag(tag, attrs)
        self.handle_endtag(tag)

    def handle_data(self, data):
        """Handle data in tags."""
        if value := data.strip():
            self.elements[-1].append(value)

    def handle_comment(self, data):
        """Handle comment string."""
        if desc := data.strip():
            self.desc.append(desc)

    def error(self, message):
        """Parser error."""
        raise NotImplementedError


def read_ptx(ptx):
    """PTX file reader.

    Parameters
    ----------
    ptx: str or pathlib.Path
        PTX file content text or file name.

    Returns
    -------
    PointingRequestMessage
        Parsed pointing request content.

    """
    content = str(ptx).strip()

    if not content.startswith('<'):
        content = Path(ptx).read_text(encoding='utf-8').strip()

    parser = PtxParser()
    parser.feed(content)

    return parser.data


# Alias `read_ptx` in `read_ptr`
read_ptr = read_ptx



"""AGM simulation module."""

from pathlib import Path

from .api import agm_api
from .results import AGMResults


def agm_simulation(metakernel, ptr, endpoint, cache=True) -> AGMResults:
    """ESA Attitude Generator Module simulation.

    Parameters
    ----------
    metakernel: str
        Metakernel id.
    ptr: str or pathlib.Path
        Pointing Timeline Request content or file.
    endpoint: str, optional
        AGM endpoint simulator. It can be an explicit API URL endpoint
        or an implicit API keyword, eg. ``'JUICE_API'``.
    cache: bool, optional
        Use cache the result of the simulation.

    Note
    ----
    Currently, AGM API requests is supported. Usage of locally installed AGM instance
    is envision in the future.

    See Also
    --------
    ptr.agm.api.agm_api, ptr.agm.AGM_CACHE

    """
    ptr_content = str(ptr).strip()

    if not ptr_content.startswith('<'):
        ptr_content = Path(ptr).read_text(encoding='utf-8').strip()

    return agm_api(metakernel, ptr_content, endpoint, cache=cache)



"""AGM results module."""

import json
import re
from html import escape
from pathlib import Path
from urllib import request

from ..datetime import dt
from ..html import HTMLCollapsible, html


class AGMResults:
    """AGM results object.

    Parameters
    ----------
    results: dict or pathlib.Path
        AGM formatted results.

    """

    def __init__(self, results):
        if isinstance(results, Path):
            with results.open('r') as fp:
                self.data = json.load(fp)

        elif isinstance(results, dict):
            self.data = results

        else:
            raise TypeError('Results must be a pathlib.Path or a dict. '
                            f'`{results.__class__.__name__}` provided.')

        # Parse the data content
        self.ptr = AGMResultsPTR(self['ptr'])
        self.status = AGMResultsStatus(self['results']['success'])

        if self.status.success:
            self.log = AGMResultsLog(self['results']['output'])
            self.ck = AGMResultsCK(self)
            self.quaternions = AGMResultsQuaternions(self['results']['quaternions'])
        else:
            self.log = AGMResultsLog(self['results']['errors'])
            self.ck = None
            self.quaternions = None

    def __str__(self):
        return json.dumps(self.data, indent=4)

    def __repr__(self):
        return f'<{self.__class__.__name__}> {self}'

    def _repr_html_(self):
        return '\n'.join([
            html(self.status),
            '<hr/>',
            html(HTMLCollapsible(
                ('Input Parameters', AGMInputParameters(self)),
                ('Output Results', AGMOutputResults(self) if self.status.success else ''),
                ('Log', self.log, '' if self.status.success else 'expand'),
            )),
        ])

    def __iter__(self):
        if self.status.success:
            return iter(self.ck)  # Load-able by spiceypy
        raise IOError('Simulation failed. No CK available')

    def __getitem__(self, item):
        try:
            return self.data[item]
        except KeyError:
            raise KeyError(f'`{item}` is not present in the results.') from None

    @property
    def success(self) -> bool:
        """Results status success flag."""
        return self.status.success


class AGMInputParameters:
    """AGM input parameters."""
    def __init__(self, res):
        self.endpoint = res['endpoint']
        self.mk = res['metakernel']
        self.ptr = res.ptr

    def __repr__(self):
        return str({
            'endpoint': self.endpoint,
            'metakernel': self.mk,
            'ptr': self.ptr,
        })

    def _repr_html_(self):
        return '\n'.join([
            '<ul style="color:#777">',
            f'<li><b>AGM Endpoint:</b> <code>{self.endpoint}</code></li>',
            f'<li><b>Metakernel:</b> <code>{self.mk}</code></li>'
            '</ul>',
            HTMLCollapsible(('PTR', self.ptr, 'expand')).html,
        ])


class AGMOutputResults:
    """AGM output results."""
    def __init__(self, res):
        self.ck = res.ck
        self.quaternions = res.quaternions

    def __repr__(self):
        return str({
            'ck': self.ck,
            'quaternions': self.quaternions,
        })

    def _repr_html_(self):
        return '\n'.join([
            '<ul style="color:#777">',
            f'<li><b>CK:</b> <code>{self.ck}</code></li>'
            '</ul>',
            HTMLCollapsible(('Quaternions', self.quaternions)).html,
        ])


class AGMResultsPTR:
    """AGM PTR object."""
    def __init__(self, ptr):
        self.ptr = ptr

    def __repr__(self):
        return self.ptr

    def _repr_html_(self):
        return f'<pre><code>{escape(self.ptr)}</code></pre>'


class AGMResultsQuaternions:
    """AGM results quaternions object."""
    HTML_MAX_LINES = 25

    def __init__(self, quaternions):
        self._raw = quaternions

        self.data = [
            [dt(time), (float(qx), float(qy), float(qz), float(qw))]
            for line in quaternions.splitlines()[1:]  # Skip header
            for time, qx, qy, qz, qw in [line.split(',')]
        ]

    def __repr__(self):
        return '\n'.join(
            f'{t} | {x:.16f} | {y:0.16f} | {z:.16f} | {w:.16f}'
            for t, (x, y, z, w) in self.data
        )

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        return iter(self.data)

    def __getitem__(self, item):
        return self.data[item]

    def _repr_html_(self):
        header = '<tr><th>time</th><th>qx</th><th>qy</th><th>qz</th><th>qw</tr>'
        lines = ''.join([
            f'<tr><td><em>{t}</em></td><td>{x}</td>'
            f'<td>{y}</td><td>{z}</td><td>{w}</td></tr>'
            for t, (x, y, z, w) in self.data[:self.HTML_MAX_LINES]
        ])

        if len(self) > self.HTML_MAX_LINES:
            lines += (
                '<tfoot>'
                '<tr><td>&hellip;</td><td>&hellip;</td><td>&hellip;</td>'
                '<td>&hellip;</td><td>&hellip;</td></tr>'
                '<tr><td colspan="5" style="text-align: center;">'
                '<em>Use <code>print()</code> to display all elements.</em>'
                '</td></tr></tfoot>'
            )

        return '<table>' + header + lines + '</table>'


class AGMResultsCK:
    """AGM results CK object."""
    def __init__(self, res):
        if cache := res['cache']:
            self.fname = Path(cache['location']) / (cache['md5_hash'] + '.ck')
        else:
            key = res['results']['ck_path'].split('/')[1]
            self.fname = Path(f'AGM_{key}.ck')

        if '://' in (url := res['endpoint']):
            protocol, uri = url.split('://')
            self.url = protocol + '://' + uri.split('/')[0] + res['results']['ck_path']
        else:
            self.url = None

    def __str__(self):
        return str(self.fname) if self.fname.exists() else self.url

    def __repr__(self):
        return f'<{self.__class__.__name__}> {self}'

    def __iter__(self):
        if not self.fname.exists():
            self.download()

        yield str(self.fname)  # Load-able by spiceypy

    def download(self):
        """Download CK file."""
        if self.url is None:
            raise ValueError('Endpoint is not a remote URL.')

        # Download CK file
        request.urlretrieve(self.url, self.fname)

        return self.fname

    def save(self, fout, overwrite=False):
        """Save CK file into an new location."""
        fout = Path(fout)

        if not overwrite and fout.exists():
            raise FileExistsError(fout)

        for f in self:
            ck_content = Path(f).read_bytes()
            fout.write_bytes(ck_content)

        self.fname = fout
        return self.fname


class AGMResultsLog:
    """AGM results log object."""
    HTML_MAX_LINES = 25
    line = re.compile(r'\[(\w+)\]\s*<([\w\s><]+)>\s*([\w\s\/\.\:+\-_]+)')

    def __init__(self, log):
        self.log = [
            (flag, tag.replace('AGE> <', ''), msg)
            for line in log.split('\n')
            for flag, tag, msg in self.line.findall(line)
        ]

    def __repr__(self):
        return '\n'.join(' | '.join(line) for line in self.log)

    def _repr_html_(self):
        lines = ''.join([
            f'<tr><td>{flag}</td><td>{tag}</td>'
            f'<td style="text-align: left;"><em>{escape(msg)}</em></td></tr>'
            for flag, tag, msg in self.log[:self.HTML_MAX_LINES]
        ])

        if len(self) > self.HTML_MAX_LINES:
            lines += (
                '<tfoot>'
                '<tr><td>&hellip;</td><td>&hellip;</td>'
                '<td style="text-align: left;">&hellip;</td></tr>'
                '<tr><td colspan="3" style="text-align: center;">'
                '<em>Use <code>print()</code> to display all elements.</em>'
                '</td></tr></tfoot>'
            )

        return '<table>' + lines + '</table>'

    def __len__(self):
        return len(self.log)

    def __iter__(self):
        return iter(self.log)

    def __getitem__(self, item):
        return self.log[item]


class AGMResultsStatus:
    """AGM results status object"""
    def __init__(self, success):
        self.success = success

    def __repr__(self):
        return 'Success' if self.success else 'Failure'

    def _repr_html_(self):
        return (
            f'<p><span style="color: {self.color}">{self.symbol}</span> '
            f'<b>{self}</b></p>'
        )

    @property
    def failure(self):
        """Failure flag."""
        return not self.success

    @property
    def symbol(self):
        """Status symbol."""
        return '✔' if self.success else '✘'

    @property
    def color(self):
        """Status color."""
        return '#2ca02c' if self.success else '#d62728'



"""PTR metadata module."""

from pathlib import Path

from .element import Element


class Comment(Element):
    """PTR metadata comment element.

    Parameters
    ----------
    text: str
        Comment message.
    description: str or list, optional
        Comment description, put as a xml-comment on top of the element.

    """

    def __init__(self, text, description=None):
        super().__init__('comment', str(text), description=description)


class File(Element):
    """PTR include file comment element.

    Parameters
    ----------
    fname: str
        Filename to include message.
    description: str or list, optional
        Element description, put as a xml-comment on top of the element.

    """

    def __init__(self, text, description=None):
        super().__init__('include', href=str(text), description=description)


class Metadata(Element):
    """PTR metadata element.

    Parameters
    ----------
    *comments: str, list, tuple or ptr.Element
        Metadata comments.
    description: str or list, optional
        Element description, put as a xml-comment on top of the element.

    Note
    ----
    If the comment provided is a pathlib.Path it
    will be converted as a `File` element.

    If the comment provided is not an `Element` object
    it will be formatted as an `Comment` element.

    """

    def __init__(self, *comments, description=None):
        super().__init__('metadata', *comments, description=description)

    def append(self, element):
        """Append new metadata."""
        if isinstance(element, (list, tuple)):
            for el in element:
                self.append(el)
            return self

        if isinstance(element, Path):
            element = File(element)

        elif element and not isinstance(element, (Element, dict)):
            element = Comment(element)

        return super().append(element)


class WithMetadata:
    """Add Metadata properties to Element objects."""

    def insert(self, index, element):
        """Insert element before a given index."""
        raise NotImplementedError

    def add_metadata(self, *comments):
        """Append comment(s) to the element."""
        try:
            meta = self['metadata']
        except KeyError:
            meta = Metadata()
            self.insert(0, meta)

        meta.append(comments)


class ElementMetadata(Element, WithMetadata):
    """Element with metadata properties."""

    def __init__(self, tag, *elements, metadata=None, description=None, **attrs):
        super().__init__(tag, description=description, **attrs)
        if metadata:
            self.add_metadata(metadata)

        for element in elements:
            self.append(element)



"""PTR timeline module."""

from warnings import warn

from .block.slew import SLEW_BLOCK
from .block.utils import gap, insort, is_block
from .datetime import dt, td
from .element import XML, Element


class Timeline(Element):
    """Planning Timeline Request element.

    Parameters
    ----------
    *blocks: ObsBlock, optional
        Block element(s) (default: None).
    frame: str, optional
        Timeline reference frame (default: 'SC').
    description: str or list, optional
        Timeline description, put as a xml-comment on top of the element.
    **attrs:
        Timeline attributes keywords.

    Note
    ----
    The blocks are automatically sorted. Overlaps
    is not allowed and will raise a BlockOverlapError.
    Slew blocks are automatically appended to fill
    the gaps between the blocks. They should not be
    included in the blocks input list (it will return
    a TypeError).

    """

    def __init__(self, *blocks, frame='SC', **attrs):
        super().__init__('timeline', frame=frame, **attrs)

        for block in blocks:
            self.append(block)

    @property
    def start(self):
        """Timeline start time."""
        return self[0].start if self else None

    @property
    def end(self):
        """Timeline end time."""
        return self[-1].end if self else None

    @property
    def duration(self):
        """Timeline duration."""
        return self.end - self.start if self else None

    @property
    def xml(self):
        """Element XML object."""
        xml = XML.createElement(self.tag)

        # Add element attributes
        for key, value in self.attrs.items():
            if value is not None:
                xml.setAttribute(key, str(value))

        latest = False
        for block in self:
            # Add a slew between blocks only when required
            if latest and gap(latest, block):
                xml.appendChild(SLEW_BLOCK.xml)

            # Add child description if present
            for desc in block.xml_desc:
                xml.appendChild(desc)

            xml.appendChild(block.xml)
            latest = block

        return xml

    @is_block
    def append(self, element):
        """Append a new block to the timeline.

        Warning
        -------
        The new block will be added chronologically in
        the list of elements. It must fit between the other
        blocks.

        """
        insort(self._els, element)
        element.link_timeline(self)  # Link this timeline to this block element

        return self

    def insert(self, _, element):
        """Disable index insertion."""
        warn('Insertion is always performed chronologically. '
             'You should use `.append()` instead.',
             SyntaxWarning, stacklevel=2)

        return self.append(element)

    def pop(self, key=-1):
        """Remove block from timeline."""
        block = super().pop(key)
        block.unlink_timeline(self)
        return block

    def sort(self):
        """Sort blocks."""
        self._els.sort()

    def offset(self, offset, *, ref='start'):
        """Offset timeline blocks globally.

        Parameters
        ----------
        offset: str datetime.timedelta or datetime.datetime
            Global or relative offset.

        ref: str, optional
            Boundary reference for relative offset (datetime).
            Only ``start|end|center`` are accepted

        Raise
        -----
        KeyError
            If the reference keyword is invalid.

        Note
        ----
        This operation does not change the duration of the timeline.

        """
        try:
            delta = td(offset)
        except ValueError:
            t = dt(offset)
            if ref == 'start':
                delta = t - self.start.datetime
            elif ref == 'end':
                delta = t - self.end.datetime
            elif ref == 'center':
                delta = t - (self.start + self.duration / 2)
            else:
                raise KeyError('For relative offset, the ref keyword must be '
                               'in `start|end|center`.') from None

        # Move the blocks from the end when delta > 0 to avoid collisions
        for block in self if delta.total_seconds() <= 0 else reversed(self):
            block.offset(delta)

        return self



"""Pointing Request Message module."""

from .agm import AGMResults, agm_simulation
from .element import XML, Element
from .metadata import Comment, ElementMetadata
from .timeline import Timeline


class PointingRequestMessage(Timeline):
    """Pointing Request Message element.

    Parameters
    ----------
    *blocks: ObsBlock, optional
        Block element(s).
    header: str, ptr.Element or list, optional
        Header comments.
    description: str or list, optional
        Message top comment description.
    seg_name: str, optional
        Segment name attribute.
    seg_metadata: str or list, optional
        Segment metadata comments.
    frame: str, optional
        Timeline reference frame (default: 'SC').
    **attrs:
        Timeline attributes keywords.

    Warning
    -------
    Only 1 segment is accepted for PTR.
    Support for PTS is out of scope for now.

    """

    def __init__(self, *blocks, header=None, description=None,
                 seg_name=None, seg_metadata=None, frame='SC', **attrs):
        super().__init__(*blocks, frame=frame, description=description, **attrs)

        self.header = Element('header')
        self.add_header(header)

        self.seg_name = seg_name
        self.seg_metadata = seg_metadata

    def add_header(self, element):
        """Append elements to the header."""
        if isinstance(element, (Element, dict)):
            self.header.append(element)

        elif isinstance(element, (list, tuple)):
            for el in element:
                self.add_header(el)

        elif element:
            # Add single value into a Comment element
            self.header.append(Comment(element))

    @property
    def xml(self):
        """PRM XML representation."""
        prm = XML.createElement('prm')

        # Add header only if present
        if self.header:
            prm.appendChild(self.header.xml)

        # Wrap prm > body > segment > data > timeline
        body = Element('body').xml
        seg = ElementMetadata('segment',
                              name=self.seg_name, metadata=self.seg_metadata).xml
        data = Element('data').xml
        timeline = super().xml

        data.appendChild(timeline)
        seg.appendChild(data)
        body.appendChild(seg)
        prm.appendChild(body)

        return prm

    def simulate(self, metakernel, agm_endpoint, cache=True) -> AGMResults:
        """Simulate Pointing Request with the Attitude Generator Module.

        Parameters
        ----------
        metakernel: str
            Baseline metakernel for AGM input
        agm_endpoint: str
            AGM endpoint.
        cache: bool, optional
            Use cache the response if present locally.

        Returns
        -------
        AGMResults
            AGM simulation results. If the simulation succeeded, the resulting CK
            can be loading into spiceypy.

        """
        return agm_simulation(metakernel, str(self), agm_endpoint, cache=cache)
